FreeRefills = LibStub("AceAddon-3.0"):NewAddon("FreeRefills","AceEvent-3.0","AceConsole-3.0")
local addon = FreeRefills

local L = LibStub("AceLocale-3.0"):GetLocale("FreeRefills")
local LBU = LibStub("LibBagUtils-1.0")


local floor,ceil,min,max = floor,ceil,min,max
local select,type,pairs,next = select,type,pairs,next
local tonumber,tostring = tonumber,tostring
local strmatch,strfind,format = strmatch,strfind,format
local assert=assert
-- GLOBALS: LibStub
-- GLOBALS: BANK_CONTAINER
-- GLOBALS: GetItemInfo,GetItemCount,GetContainerNumFreeSlots,PickupContainerItem,SplitContainerItem
-- GLOBALS: GetMerchantNumItems,GetMerchantItemInfo,BuyMerchantItem



function addon:OnInitialize()
	local defaults = {
		profile = {
			overstock = false,
			merchant = true,
			bank = true,
			refills = {},
		}
	}

	self.db = LibStub("AceDB-3.0"):New("FreeRefillsDB", defaults --[[, char default]])

	local options
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("FreeRefills",function()
		if not options then	-- avoid table creation until necessary
			options = {	
				type = 'group',
				name="FreeRefills",	-- displayname in GUIs
				childGroups="tab",
				
				handler = self,
				set="SetDB",
				get="GetDB",
				
				args = {
					cmdlinetools  = {
						type="group",
						name="cmdlinetools",
						hidden = true,
						cmdHidden = false,
						inline = true,
						args = {
							add = {
								order=1,
								name = L["Add"], type = 'input',
							    desc = L["Add an item to the refill list. (Usage: /freerefills add [Item Link] #)"],
								get = false,
							    set = "AddRefillItem",
								usage = L["[Item Link] <number of items>"],
								guiHidden = true
							},
							addstack = {
								order=2,
								name = L["Add Stack"], type = 'input',
							    desc = L["Add an item to the refill list by stack. (Usage: /freerefills addstack [Item Link] #)"],
							    get = false,
								set = "AddRefillItemStack",
								usage = L["[Item Link] <number of stacks>"],
								guiHidden = true
							},
							del = {
								order=3,
								name = L["Remove"], type = 'input',
							    desc = L["Remove an item from the refill list. (Usage: /freerefills del [Item Link])"],
								get = false,
							    set = "RemoveRefillItem",
								usage = L["[Item Link]"],
								guiHidden = true
							},
							list = {
								order=4,
								name = L["List"], type = 'execute',
							    desc = L["List the items in the refill list."],
							    func = "List",
							},
						},
					},
					gui = {
						cmdHidden = true,
						type="description",
						name = "Use the command line to manage the list of items to be refilled. Use \"/freerefills help\"",
					},
					overstock = {
						order=200,
						name = L["Overstock"], type = 'toggle',
						desc = L["Toggle Overstocking.  If an item is sold in stacks and cannot meet your exact requested amount, you can opt to buy slightly more (Overstock)."],
					},
					merchant = {
						order=201,
						name = L["Process Merchant"], type = 'toggle',
						desc = L["Toggle Merchant Looting."],
					},
					bank = {
						order=202,
						name = L["Process Bank"], type = 'toggle',
						desc = L["Toggle Bank Looting."],
					},
					reset = {
						order=299,
						name = L["Clear"], type = 'execute',
					    desc = L["Clear the refill list completely."],
					    func = "ClearRefills",
					},
					profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db),
				}
			}
		end
		
		return options
	end)
	
	self:RegisterChatCommand("freerefills", "OnChatCommand", true)
	self:RegisterChatCommand("fr", "OnChatCommand", true)
end

addon.onSets={}
function addon:SetDB(info,v)
	local name = info[#info]
	self.db.profile[name]=v
	if self.onSets[name] then
		self[self.onSets[name]](self,v)
	end
end

function addon:GetDB(info)
	return self.db.profile[info[#info]]
end

function addon:OnChatCommand(input)
	LibStub("AceConfigCmd-3.0"):HandleCommand("freerefills", "FreeRefills", input)
end

addon.onSets["merchant"]="OnSet_merchant"
function addon:OnSet_merchant(v)
	self[v and "RegisterEvent" or "UnregisterEvent"](self,"MERCHANT_SHOW")
	self:Print("Will",v and "now" or "no longer","refill from merchants.")
end

addon.onSets["bank"]="OnSet_bank"
function addon:OnSet_bank(v)
	self[v and "RegisterEvent" or "UnregisterEvent"](self,"BANKFRAME_OPENED")
	self:Print("Will",v and "now" or "no longer","refill from the bank.")
end

addon.onSets["overstock"]="OnSet_overstock"
function addon:OnSet_overstock(v)
	self:Print("Will",v and "now" or "no longer","overstock.")
end


-- Okay, this is the register-for-an-event-inside-being-LoDed-from-the-same-event bullshit again
-- See http://code.google.com/p/tullamods/issues/detail?id=325
-- and http://www.wowace.com/addons/free-refills/tickets/4-conflict-with-combuctor/
local frame=CreateFrame("Frame", "FreeRefillsInit")
frame:Hide()
frame:SetScript("OnUpdate", function()
	if addon.db.profile.merchant then
		addon:RegisterEvent("MERCHANT_SHOW")
	end
	if addon.db.profile.bank then
		addon:RegisterEvent("BANKFRAME_OPENED")
	end
	frame:Hide()
end)

function addon:OnEnable()
	frame:Show()
	
	-- In case we get loaded by AddonLoader... (in which case we don't get to see the events ourself)
	if self.db.profile.bank and select(2,GetContainerNumFreeSlots(BANK_CONTAINER)) then
		self:BANKFRAME_OPENED()
	end
	if self.db.profile.merchant and GetMerchantItemInfo(1) then
		self:MERCHANT_SHOW()
	end
		
end

function addon:debug(msg)
	--DBG self:Print("|cff0099CC"..msg.."|r")
end

--[[--------------------------------------------------------------------------------
  Merchant Processing
-----------------------------------------------------------------------------------]]

function addon:MERCHANT_SHOW()
	self:debug("Merchant Opened")
	for itemID, amount in pairs(self.db.profile.refills) do
		local merchSlot = self:ItemInStock(itemID)
		if merchSlot then
			self:debug("Item Found in Merchant Slot #"..merchSlot)
			local amountInBags = GetItemCount(itemID)
			
			if amountInBags >= amount then
				self:debug("Already have enough in bags.")
			else
				local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture
					= GetItemInfo(itemID)

				if not itemStackCount then
					self:Print("Err, something odd happened: the item '"..tostring(itemID).."' did not have a stack size?. This could be caused by you having just patched and the WoW item cache not yet having been updated (and in that case is perfectly normal). Try showing the item's tooltip to force an update from the server.")
					return
				end

				local name, texture, price, quantity, numAvailable, isUsable, extendedCost 
					= GetMerchantItemInfo(merchSlot)
				if numAvailable<0 then numAvailable=9e9 end
				
				local toBuy = (amount-amountInBags)/quantity
				if self.db.profile.overstock then
					toBuy = ceil(toBuy)
				else
					toBuy = floor(toBuy)
				end
				
				toBuy=min(toBuy,numAvailable)

				local fullStackQty = floor(itemStackCount / quantity)
				assert(fullStackQty>=1)
				
				while toBuy>0 do
					local now = min(toBuy,fullStackQty)
					self:Print(format("  %2u x %s", now*quantity, itemLink))
					BuyMerchantItem(merchSlot, now)
					toBuy=toBuy-now
				end
			end
		end
	end
end

function addon:ItemInStock(itemLink)
	self:debug("Searching for Item: "..itemLink)
	local itemName = GetItemInfo(itemLink)
	local numItems = GetMerchantNumItems()
	for i = 1, numItems do
		if GetMerchantItemInfo(i) == itemName then		-- compares names
			return i
		end
	end
end



--[[--------------------------------------------------------------------------------
  Bank Processing
-----------------------------------------------------------------------------------]]

function addon:BANKFRAME_OPENED()
	self:debug("Bank Opened")
	for itemID,amount in pairs(self.db.profile.refills) do
		self:debug(itemID..": "..amount)
		local amountInBags = GetItemCount(itemID)
		local amountInBank = GetItemCount(itemID, true) - amountInBags
		if amountInBank > 0 then
			local amountNeeded = amount - amountInBags
			if amountNeeded<=0 then
				self:debug("Already have enough in bags.")
			else
				if amountNeeded > amountInBank then
					amountNeeded = amountInBank 
				end
				self:debug(amountNeeded.." item(s) to be looted.")
				self:LootBank(itemID, min(amountNeeded, amountInBank))
			end
		end
	end
end

local function linkDecode(link)
	if link then
		return select(3, link:find("Hitem[^|]+|h%[([^[]+)%]"))
	end
end

function addon:LootBank(itemID, amountNeeded)
	self:debug("Bank Looting Logic: "..itemID..": "..amountNeeded)
	
	while amountNeeded>0 do
		local bag,slot,size = LBU:FindSmallestStack("BANK", tonumber(itemID) or tonumber(strmatch(itemID,"item:([0-9]+)")))  -- won't find locked slots
		if not bag then
			self:debug("  - no (more) in bank.")
			break
		end
		self:debug("  found " .. size .. " in "..bag..","..slot)
		
		if size<=amountNeeded then
			self:debug("      looting entire stack")
			-- Don't use this or LBU doesn't know the slot is taken: UseContainerItem(bag,slot)
			PickupContainerItem(bag,slot)
			LBU:PutItem("BAGS", size)
			amountNeeded = amountNeeded - size
		else
			self:debug("      splitting off "..amountNeeded)
			SplitContainerItem(bag, slot, amountNeeded)
			LBU:PutItem("BAGS", amountNeeded)
			amountNeeded = 0
		end
		
	end

end

--[[--------------------------------------------------------------------------------
  Command Handlers
-----------------------------------------------------------------------------------]]

local function setrefill(link,num)
	-- itemstrings now have the player's CURRENT level in them, so we have to take the itemstring apart for matching
	local substr = strmatch(link, "item:%d+:") or "["..link.."]"
	for k,v in pairs(addon.db.profile.refills) do
		if strfind(k, substr, 1, true) then
			addon.db.profile.refills[k] = num
			return k
		end
	end
	if num then
		addon.db.profile.refills[link] = num
	end
end

function addon:AddRefillItem(info, msg)
	--add an item to the refill list
	local link, num = strmatch(msg,"^%s*(|c.+|Hitem.+)%s(%d+)%s*$")
	if not link or not num then
		self:Print(L["Invalid input; input must be '<item> #'."])
		return
	end
	num = tonumber(num)
	setrefill(link,num)
	self:Print(L["Added %s - Amount:%d"]:format(link, num))
end

local function GetStackSize(item)
	local itemStack = select(8, GetItemInfo(item))
	return itemStack
end

function addon:AddRefillItemStack(info, msg)
	--add an item to the refill list by stack
	local link, num = strmatch(msg,"^%s*(|c.+|Hitem.+)%s(%d+)%s*$")
	if not link or not num then
		self:Print(L["Invalid input; input must be '<item> #'."])
		return
	end
	num = tonumber(num)
	num = num * GetStackSize(link)
	setrefill(link,num)
	self:Print(L["Added %s - Amount:%d"]:format(link, num))
end

function addon:RemoveRefillItem(info, msg)
	--remove an item from the refill list
	local item = setrefill(msg, nil)
	if item then
		self:Print(L["Removed %s."]:format(item))
	else
		self:Print(L["Could not find '%s'."]:format(msg))
	end
end

function addon:ClearRefills()
	--clear the refill list
	self.db.profile.refills = {}
	self:Print(L["Item List Cleared"])
end

function addon:List()
	--report back which items are currently on the refill list, and their quanitity
	self:Print(L["Refill List: "])
	for itemID,amount in pairs(self.db.profile.refills) do
		self:Print("     "..itemID.." ("..amount..")")
	end
	if not next(self.db.profile.refills) then 
		self:Print(L["     None"]) 
	end
end


